#include <stddef.h>

#include "log.h"
#include "common/xmalloc.h"
#include "bisect.h"

struct trace_flag *tf_bisect(struct bisect_meta *meta, pid_t key)
{
	struct trace_flag **tfs = meta->data;
	int lo = 0, hi = meta->used, mid;

	if (meta->used <= 0)
		return NULL;

	while (lo < hi) {
		mid = (int)((lo + hi) / 2);
		if (tfs[mid]->key == key) {
			return tfs[mid];
		} else if (tfs[mid]->key > key) {
			hi = mid;
		} else {
			lo = mid + 1;
		}
	}

	return NULL;
}

/* used in cr-restore */
struct trace_flag *tf_insert(struct bisect_meta *meta, pid_t key)
{
	struct trace_flag **tfs = meta->data;
	struct trace_flag *tf = &((struct trace_flag *)meta->__data)[meta->used];
	int i = 0, j = 0;

	if (meta->used == meta->size)
		return NULL;

	for (i = 0; i < meta->used; i++) {
		if (tfs[i]->key >= key) /* impossible condition: `tfs[i]->key == key` */
			break;
	}

	j = meta->used;
	meta->used += 1;

	while (j > i) {
		tfs[j] = tfs[j-1];
		j -= 1;
	}

	tfs[i] = tf;
	tf->key = key;

	return tf;
}

int tf_create(struct bisect_meta *meta, int len)
{
	struct trace_flag *tfs;
	struct trace_flag **tfs_ptr;

	tfs = xzalloc(sizeof(*tfs) * len);
	if (tfs == NULL)
		return -1;

	tfs_ptr = xmalloc(sizeof(*tfs_ptr) * len);
	if (tfs_ptr == NULL)
		goto err;

	meta->size = len;
	meta->used = 0;
	meta->__data = tfs;
	meta->data = tfs_ptr;

	return 0;
err:
	xfree(tfs);
	return -1;
}

void tf_destroy(struct bisect_meta *meta)
{
	xfree(meta->__data);
	xfree(meta->data);
}

void tf_clear(struct bisect_meta *meta)
{
	meta->used = 0;
	__builtin_memset(meta->data, 0, sizeof(struct trace_flag **)*meta->size);
}
