#ifndef __CR_NFTABLES_H__
#define __CR_NFTABLES_H__

#include <libmnl/libmnl.h>

#include <libnftnl/table.h>
#include <libnftnl/chain.h>
#include <libnftnl/set.h>
#include <libnftnl/rule.h>
#include <libnftnl/expr.h>

#define construct_buf(buf, type, family, flags, seq, payload, cb_prefix)	\
	({	\
		struct nlmsghdr *_nlh;	\
		\
		_nlh = nftnl_##cb_prefix##_nlmsg_build_hdr((buf),	\
					(type), (family), (flags), (seq));	\
		nftnl_##cb_prefix##_nlmsg_build_payload(_nlh, (payload));	\
		nftnl_##cb_prefix##_free((payload));	\
		_nlh;	\
	})

#define construct_table_buf(buf, type, family, flags, seq, payload)	\
		construct_buf((buf), (type), (family), (flags), (seq),	\
				(payload), table)

#define construct_chain_buf(buf, type, family, flags, seq, payload)	\
		construct_buf((buf), (type), (family), (flags), (seq),	\
				(payload), chain)

#define construct_batch(batch, type, family, flags, seq, payload, cb_prefix)	\
	{	\
		struct nlmsghdr *_nlh;	\
		\
		_nlh = nftnl_##cb_prefix##_nlmsg_build_hdr(	\
					mnl_nlmsg_batch_current(batch),	\
					(type), (family), (flags), (seq));	\
		nftnl_##cb_prefix##_nlmsg_build_payload(_nlh, (payload));	\
		nftnl_##cb_prefix##_free((payload));	\
		mnl_nlmsg_batch_next((batch));	\
	}

#define construct_table_batch(batch, type, family, flags, seq, payload)	\
		construct_batch((batch), (type), (family), (flags), (seq),	\
				(payload), table)

#define construct_chain_batch(batch, type, family, flags, seq, payload)	\
		construct_batch((batch), (type), (family), (flags), (seq),	\
				(payload), chain)

#define construct_set_batch(batch, type, family, flags, seq, payload)	\
		construct_batch((batch), (type), (family), (flags), (seq),	\
				(payload), set)

#define construct_rule_batch(batch, type, family, flags, seq, payload)	\
		construct_batch((batch), (type), (family), (flags), (seq),	\
				(payload), rule)

#define construct_set_elems_batch(batch, type, family, flags, seq, payload)	\
	{	\
		struct nlmsghdr *_nlh;	\
		\
		_nlh = nftnl_nlmsg_build_hdr(	\
					mnl_nlmsg_batch_current(batch),	\
					(type), (family), (flags), (seq));	\
		nftnl_set_elems_nlmsg_build_payload(_nlh, (payload));	\
		nftnl_set_free((payload));	\
		mnl_nlmsg_batch_next((batch));	\
	}

#define TABLE_NAME	"filter"
#define INPUT_CHAIN_NAME	"criu-input"
#define OUTPUT_CHAIN_NAME	"criu-output"
#define INPUT_IPV4_SET_NAME	"criu-input-ipv4-blacklist-%d"
#define INPUT_IPV6_SET_NAME	"criu-input-ipv6-blacklist-%d"
#define OUTPUT_IPV4_SET_NAME	"criu-output-ipv4-blacklist-%d"
#define OUTPUT_IPV6_SET_NAME	"criu-output-ipv6-blacklist-%d"

/* set key type, see nftables/include/datatypes.h
 * The rule of the datatype calculation:
 *   Each type occupies 6 bits, type:
 *   - ipaddr: 7, 4 bytes
 *   - ip6addr: 8, 16 types
 *   - inet_service: 13, 2 bytes (pading to 4 bytes)
 *
 *   0x1cd1cd: 0b 000111 001101 000111 001101
 *   0x20d20d: 0b 001000 001101 001000 001101
 */
#define INET_SERVICE_LEN	2
#define IPADDR_LEN	4
#define IP6ADDR_LEN	16
#define div_round_up(n, d)	(((n) + (d) - 1) / (d))

#define IPv4_KEY_TYPE	0x1cd1cd
#define IPv4_KEY_LEN	div_round_up(IPADDR_LEN + INET_SERVICE_LEN, 4) * 4 * 2
#define IPv6_KEY_TYPE	0x20d20d
#define IPv6_KEY_LEN	div_round_up(IP6ADDR_LEN + INET_SERVICE_LEN, 4) * 4 * 2

struct mnl_params {
	struct mnl_socket *nl;
	char *buf;
	struct mnl_nlmsg_batch *batch;
	uint32_t seq;
};

typedef struct nlmsghdr * (*buf_func_t)(struct mnl_params *mnl_params, void *args);
typedef int (*batch_func_t)(struct mnl_params *mnl_params, void *args);
typedef int (*mnl_func_t)(struct mnl_params *mnl, batch_func_t cb, void *args);

struct mnl_cb_params {
	pid_t tree_id;
	bool create;
	bool ipv6;
};

int mnl_sendmsg(batch_func_t batch_cb, void *args);
int mnl_common(mnl_func_t mnl_cb, void *arg1, void *arg2);
int mnl_batch_send_and_recv(struct mnl_params *mnl_params, batch_func_t cb, void *args, int *result);
int mnl_buf_send_and_recv(struct mnl_params *mnl_params, buf_func_t cb, void *args, int *result);

struct nft_chain_params {
	char *name;
	uint32_t hooknum;
	char *type;
	uint32_t prio;
	uint32_t policy;
};

struct nft_set_params {
	char name[128];
	uint32_t id;
	uint32_t datatype;
	uint32_t key_len;
};

struct nft_rule_params {
	char *chain_name;
	char set_name[128];
	uint32_t mark;
	uint16_t mark_op;
	uint32_t nfproto;
	uint8_t l4proto;
	unsigned int stmt;
	bool ipv6;
};

struct nft_set_elem_params {
	char set_name[128];
	char data[40];
	size_t data_len;
};

struct nf_conn_params {
	uint8_t family;
	uint32_t *src_addr;
	uint16_t src_port;
	uint32_t *dst_addr;
	uint16_t dst_port;
	bool lock;
	pid_t tree_id;
};

struct inet_sk_desc;
int nft_connection_switch(struct inet_sk_desc *sk, bool lock, pid_t tree_id);

#endif /* __CR_NFTABLES_H__ */
