#ifndef __CRIU_PIN_MEM_H__
#define __CRIU_PIN_MEM_H__

#include <stdbool.h>

#include "vma.pb-c.h"

#if __has_include("linux/pin_memory.h")
# define CONFIG_PID_RESERVE
# include <linux/pin_memory.h>
#else

#define PIN_MEM_MAGIC		0x59
#define _SET_PIN_MEM_AREA	1
#define _CLEAR_PIN_MEM_AREA	2
#define _REMAP_PIN_MEM_AREA	3
#define _DUMP_SEPCIAL_PAGES	6
#define _RETORE_SEPCIAL_PAGES	7

#define SET_PIN_MEM_AREA	_IOW(PIN_MEM_MAGIC, _SET_PIN_MEM_AREA, struct pin_mem_area_set)
#define CLEAR_PIN_MEM_AREA	_IOW(PIN_MEM_MAGIC, _CLEAR_PIN_MEM_AREA, int)
#define REMAP_PIN_MEM_AREA	_IOW(PIN_MEM_MAGIC, _REMAP_PIN_MEM_AREA, int)
#define DUMP_SPECIAL_PAGES	_IOW(PIN_MEM_MAGIC, _DUMP_SEPCIAL_PAGES, int)
#define RETORE_SPECIAL_PAGES	_IOW(PIN_MEM_MAGIC, _RETORE_SEPCIAL_PAGES, int)

#define MAX_PIN_MEM_AREA_NUM		16

struct _pin_mem_area {
	unsigned long virt_start;
	unsigned long virt_end;
};

struct pin_mem_area_set {
	unsigned int pid;
	unsigned int area_num;
	struct _pin_mem_area mem_area[MAX_PIN_MEM_AREA_NUM];
};

#define _SET_FORK_PID	8
#define SET_FORK_PID		_IOW(PIN_MEM_MAGIC, _SET_FORK_PID, int)

#define _INIT_PAGEMAP_READ	5
#define INIT_PAGEMAP_READ	_IOW(PIN_MEM_MAGIC, _INIT_PAGEMAP_READ, int)

#endif /* __has_include("linux/pin_memory.h") */

#define PIN_MEM_FILE	"/dev/pinmem"
#define ONCE_PIN_MEM_SIZE_LIMIT		(32 * 1024 * 1024)

bool should_pin_vmae(VmaEntry *vmae);
int pin_vmae(VmaEntry *vmae, struct pstree_item *item);
int dump_task_special_pages(int pid);
int restore_task_special_pages(int pid);
int clear_pin_mem(int pid);
int init_pagemap_read(int para);

#endif /* __CRIU_PIN_MEM_H__ */
