#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <linux/limits.h>

#include "log.h"
#include "cr_options.h"
#include "util.h"
#include "sk-inet.h"
#include "reserved-ports.h"

#include "common/list.h"

static char* reserved_ports;
static int reserved_ports_num;
extern struct list_head cpt_tcp_repair_sockets;

void read_reserved_ports(char *path)
{
	FILE *file = NULL;
	char *ch = NULL;
	size_t size = 0;

	if (reserved_ports) {
		free(reserved_ports);
		reserved_ports = NULL;
	}

	file = fopen(path, "r");
	if (!file) {
		pr_err("Cannot fopen %s\n", path);
		return;
	}

	if (getline(&reserved_ports, &size, file) <= 0)
		pr_err("Cannot getline from %s\n", path);
	fclose(file);

	if (!reserved_ports)
		return;

	ch = strstr(reserved_ports, "\n");
	if (ch)
		*ch = '\0';
}

void write_reserved_ports(char *path)
{
	int fd = -1;
	char buf[PATH_MAX];

	fd = open(path, O_RDWR | O_CREAT, 0640);
	if (fd < 0) {
		pr_err("Cannot open %s ret %d cwd: %s\n", path, fd, buf);
		return;
	}

	cr_system(-1, fd, -1, "/usr/bin/echo",
		  (char *[]) { "echo", reserved_ports, NULL}, 0);
	close(fd);
}

static int add_reserved_ports(struct inet_sk_desc *sk)
{
	if (reserved_ports_num >= opts.reserve_ports)
		return -1;

	if (strlen(reserved_ports) == 0)
		snprintf(reserved_ports, 6, "%u", sk->src_port);
	else
		snprintf(reserved_ports + strlen(reserved_ports), 7, ",%u", sk->src_port);
	reserved_ports_num++;

	return 0;
}

void set_reserved_ports(void)
{
	struct inet_sk_desc *sk = NULL;
	size_t size = 0;

	read_reserved_ports(RESERVED_PORTS_PATH);

	write_reserved_ports("ip_local_reserved_ports");

	size = strlen(reserved_ports) + 6 * opts.reserve_ports + 1;
	if (xrealloc_safe(&reserved_ports, size))
		exit(1);

	list_for_each_entry(sk, &cpt_tcp_repair_sockets, rlist)
		add_reserved_ports(sk);

	write_reserved_ports(RESERVED_PORTS_PATH);

	free(reserved_ports);
	reserved_ports = NULL;
}
